# demo.R
# Author: Minzhe Guo (minzhe.guo@cchmc.org)
# Description: 
#				A demonstration of using SINCERA to analyse E16.5 mouse lung single cell data.
# 				The data are RNA-seq of 148 single cells from two independent sample preparations 
#               (86 cells from sample 1 and 62 cells from sample 2) from E16.5 mouse lung and RNA
#               expression values were calculated using the FPKM.
#				Through the pipeline analysis, we distinguished major cell types of fetal mouse lung, 
#               including epithelial, endothelial, smooth muscle, pericyte, and fibroblast-like cell types, 
#               and identified cell type specific gene signatures, bioprocesses, and key regulators.
# Dependencies: 
#             SINCERA a04142015
#             E16.5.Rd
# Date: April 14, 2015

# ============================= 

#######################################################
#             Load Sincera Code and E16.5 Data        #
#######################################################

#NOTE: please make sure that demo.R, sincera.R, and E16.5.Rd are placed in the same folder and the folder is the current working directory of R

source("sincera.R")
e16.5 <- load("E16.5.Rd")

#######################################################
#      Organized the Data using ExpressionSet         #
#######################################################

pd <- new("AnnotatedDataFrame", data = cells)
fd <- new("AnnotatedDataFrame", data = genes)
ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)

#######################################################
#     Parameters for selecting analysis components    #
#######################################################

do.prefiltering = TRUE
do.clustering = TRUE
do.permutation.analysis = TRUE
do.diffexpr.test = TRUE
do.celltype.enrichment = TRUE
do.celltype.validation = TRUE
do.signature.prediction = TRUE
do.drivingforce.analysis = TRUE
do.consensus.maximization = TRUE

GENE.SYMBOL.LABEL = "SYMBOL"					    # for labeling the column encoding the gene symbol info
SAMPLE.LABEL = "SAMPLE"                        		# for labeling the column encoding the sample information
CLUSTER.LABEL = "CLUSTER"  					   		# for labeling the column encoding the results of cluster assignment
TF.LABEL = "TF"										# for labeling the column encoding the transcription factors/cofactors

TTEST.MIN.SIZE = 5                             		# in differential expression test, if both groups >= 5, used Welch's t-test; otherwise, using Wilcoxon rank sum test

COMMON.TRESHOLD=5                                   # in common gene metric, expression > COMMON.TRESHOLD will considered as expressed
COMMON.PERCENTAGE=0.8                               # in common gene metric, genes that express in >COMMON.PRECENTAGE cluster cells will be considered as a common gene shared by cluster cells.
UNIQUE.RATIO=2                                      # parameters for unique gene metric
UNIQUE.QUANTILE=0.85								# parameters for unique gene metric

EXPR.SPECIFICITY.PREFIX <- "specificity_"      		# for labeling the columns encoding the results of expression specificity calculation for each group
EXPR.ABUNDANCY.PREFIX <- "abundancy_"          		# for labeling the columns encoding the results of expression abundance calculation
PREFILTERING.PREFIX <- "use_for_analysis_"	   		# for labeling the column encoding the result of prefiltering

DIFF.EXPR.PREFIX = "test_"                     		# for labeling the columns encoding the results of differential expression test for each group
CELLTYPE.ENRICHMENT.PREFIX = "use_for_celltype_"	# for labeling the columns encoding the gene list used for cell type enrichment analysis for each group
MARKER.PREFIX = "use_as_marker_"					# for labeling the columns encoding the biomarkers for each group

COMMON.PREFIX="common_"								# for labeling the columns encoding the results of common metric for each group
UNIQUE.PREFIX="unique_"								# for labeling the columns encoding the results of unique metric for each group
TEST.STATS.METRIC.PREFIX = "test_statistic_"		# for labeling the columns encoding the results of test statistic metric for each group
SYN.SIM.PREFIX="synthetic_similiary_"				# for labeling the columns encoding the results of synthetic profile similarity metric for each group
TRAINSET.PREFIX="use_for_train_"					# for labeling the columns encoding the training instances for each group
TRAIN.CLASS.PREFIX="train_class_"					# for labeling the columns encoding the class of group specific training instances
TESTSET.PREFIX="use_for_test_"						# for labeling the columns encoding the testing instances for each group
SIGNATURE.PREFIX="sig_pred_"						# for labeling the columns encoding the results of signature prediction

TG.PREFIX="use_for_tg_" 							# for labeling the columns encoding the group specific candidate regulatory targets
TF.PREFIX="use_for_tf_"								# for labeling the columns encoding the group specific candidate transcription factors/cofactors


cat("\nE16.5 Demonstration\n\n")

#######################################################
#                Gene Pre-filtering                   #
#######################################################

if (do.prefiltering == TRUE) {

	# convert expression values out of bounds to bounds
	ES <- exprs.truncate(ES, upper.bound=NULL, lower.bound=0.01)

	# calculate per-sample specificity and abundancy scores for each profile
	ES <- prefiltering(ES, group.by="SAMPLE", groups=NULL, specificity.threshold=0.7, abundancy.threshold=5, abundancy.count=2, do.batch.analysis=TRUE, export=T)

	# select profiles that passed prefiltering filters for further analysis
	ES <- ES[which(fData(ES)[,PREFILTERING.PREFIX]==1),]
}

#######################################################
#                Cluster Assignment                   #
#######################################################

if (do.clustering == TRUE) {

	# perform gene-by-gene per-sample z-score transformation
	ESz <- normalization.zscore(ES, group.by=SAMPLE.LABEL, groups=NULL, verbose=TRUE) 

	# divide cells into distinct clusters
	# the cluster membership shall be encoded in fData(ESz)$CLUSTER
	ESz <- cluster.assignment(ESz, cluster.label=CLUSTER.LABEL, num.singleton=0, h=0.5, do.shift=TRUE, export=TRUE)
	
	# to keep the cluster ids consistent with the cluster ids used in the manuscript,
	# the cluster assignment results are applied to ESz, instead of ES
}

# perform permutation analysis
if (do.permutation.analysis == TRUE) {
	cluster.permutation.analysis(ES, group.by=CLUSTER.LABEL, n=20, method="pearson", log.base=2, verbose=T) 
}

#######################################################
#                Differential Expression              #
#######################################################

if (do.diffexpr.test == TRUE) {
	# per-cluster differential expression test
	# ttest_min_size: if the sizes of both groups are greater than 5, then t-test will be used; otherwise, wilcoxon rank sum test will be used
	ES <- diff.expr.test(ES, group.by="CLUSTER", groups=NULL, diffexpr.prefix=DIFF.EXPR.PREFIX, ttest.min.size=TTEST.MIN.SIZE, do.fdr=FALSE, verbose=T, export=T, export.components="fd")
}

#######################################################
#                Cell Type Enrichment                 #
#######################################################

if (do.celltype.enrichment == TRUE) {

	# selecting cluster specific differentially expressed genes for cell type enrichment analysis
	groups <- sort(unique(pData(ES)$CLUSTER))
	prefix=CELLTYPE.ENRICHMENT.PREFIX
	for (i in groups) {
		i.de.name <- paste(prefix, i, sep="")
		i.test.name <- paste(DIFF.EXPR.PREFIX, i, sep="")
		fData(ES)[,i.de.name] <- 0
		if (i %in% c("3")) {
			fData(ES)[which(fData(ES)[,i.test.name] < 0.03), i.de.name] <- 1 
		} else {
			fData(ES)[which(fData(ES)[,i.test.name] < 0.01), i.de.name] <- 1 
		}
	}
	
	celltype.enrichment(ES, group.by=CLUSTER.LABEL, groups=NULL, celltype.enrichment.prefix=CELLTYPE.ENRICHMENT.PREFIX, verbose=TRUE)
}


#######################################################
#                Marker-based Validation              #
#######################################################

if (do.celltype.validation == TRUE) {

	# using biomarkers to validate the cell type assignments

	ESz <- normalization.zscore(ES, group.by=SAMPLE.LABEL, groups=NULL, verbose=TRUE)  
	celltype.validation(ESz, group.by=CLUSTER.LABEL, groups=c(1,2,3,5,7,8,9), marker.prefix=MARKER.PREFIX, threshold=0)
}


#######################################################
#                 Signature Prediction                #
#######################################################

if (do.signature.prediction == TRUE) {

	ES <- signature.analysis(ES, group.by = CLUSTER.LABEL, groups=c(1,2,3,5,7,8,9),
										# training set
										trainset.prefix = TRAINSET.PREFIX,
										train.class.prefix = TRAIN.CLASS.PREFIX,
										marker.prefix=MARKER.PREFIX,
										do.cc.validation = TRUE,
										# testing set
										testset.prefix=TESTSET.PREFIX,
										# common gene metric
										common.prefix = COMMON.PREFIX,
										common.threshold=COMMON.TRESHOLD, common.percentage=COMMON.PERCENTAGE,
										# unique gene metric
										unique.prefix = UNIQUE.PREFIX,
										unique.ratio=UNIQUE.RATIO, unique.quantile=UNIQUE.QUANTILE,
										# test statistic metric
										test.statistic.metric.prefix = TEST.STATS.METRIC.PREFIX, log.base=2,
										diff.expr.prefix=DIFF.EXPR.PREFIX,
										diff.expr.threshold = 0.05,
										# synthetic profile similarity metric 
										syn.sim.prefix = SYN.SIM.PREFIX,
										signature.prefix = SIGNATURE.PREFIX,
										gene.symbol.label = GENE.SYMBOL.LABEL,
										verbose=TRUE,
										export=TRUE, export.components="fd" )
}

#######################################################
#                Driving Force Analysis               #
#######################################################

if (do.drivingforce.analysis == TRUE) {

	groups <- c(9)
	prefix.tg=TG.PREFIX
	prefix.tf=TF.PREFIX
	tf.name = TF.LABEL
	
	# selecting cluster specific differentially expressed genes as candidate regulatory targets
	# selecting cluster specific differentially expressed TFs or commonly expressed TFs as candidate TFs
	
	option = 1 # configurations for quick demonstration of driving force analysis
	# option = 2 # configurations for reproducing the driving force analysis results in the manuscript
	
	if (option == 1) {
	
		# for demonstration
		for (i in groups) {
			i.test.name <- paste(DIFF.EXPR.PREFIX, i, sep="")
			i.common.name <- paste(COMMON.PREFIX, i, sep="")
			i.tg.name <- paste(prefix.tg, i, sep="")
			i.tf.name <- paste(prefix.tf, i, sep="")
			fData(ES)[,i.tg.name] <- 0
			fData(ES)[,i.tf.name] <- 0
			
			fData(ES)[which(fData(ES)[,i.test.name] < 0.0001), i.tg.name] <- 1
			fData(ES)[which(fData(ES)[,i.test.name] < 0.005), i.tf.name] <- 1
			#fData(ES)[which(fData(ES)[,i.common.name] == 1), i.tf.name] <- 1
			fData(ES)[which(fData(ES)[,tf.name] == 0), i.tf.name] <- 0	
		}
	
	} else if (option == 2) {

		# configurations for reproducing the driving force analysis results in the manuscript
		# this would take about 1 hour for running
		for (i in groups) {
			i.test.name <- paste(DIFF.EXPR.PREFIX, i, sep="")
			i.common.name <- paste(COMMON.PREFIX, i, sep="")
			i.tg.name <- paste(prefix.tg, i, sep="")
			i.tf.name <- paste(prefix.tf, i, sep="")
			fData(ES)[,i.tg.name] <- 0
			fData(ES)[,i.tf.name] <- 0
			if (i %in% c("3")) {
				fData(ES)[which(fData(ES)[,i.test.name] < 0.03), i.tg.name] <- 1
				fData(ES)[which(fData(ES)[,i.test.name] < 0.05), i.tf.name] <- 1
				fData(ES)[which(fData(ES)[,i.common.name] == 1), i.tf.name] <- 1
				fData(ES)[which(fData(ES)[,tf.name] == 0), i.tf.name] <- 0			
			} else {
				fData(ES)[which(fData(ES)[,i.test.name] < 0.01), i.tg.name] <- 1
				fData(ES)[which(fData(ES)[,i.test.name] < 0.05), i.tf.name] <- 1
				fData(ES)[which(fData(ES)[,i.common.name] == 1), i.tf.name] <- 1
				fData(ES)[which(fData(ES)[,tf.name] == 0), i.tf.name] <- 0	
			}
		}
	}
	
	driving.force.analysis(ES, group.by=CLUSTER.LABEL, groups=c(9), unique.by=GENE.SYMBOL.LABEL)
}

################################################################################
#      Consensus Maximization for Nkx2-1 Regulatory Targets Prediction         #
################################################################################

if (do.consensus.maximization == TRUE) {
	consensus_maximization(data4cm)
}

cat("\n\nDemonstration completed\n\n")
