# SINCERA: A Computational Pipeline for Single Cell RNA-seq Profiling Analysis

# ==============

# Copyright (C) 2015  Minzhe Guo (minzhe.guo@cchmc.org)

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# ================

# Author: Minzhe Guo (minzhe.guo@cchmc.org)
# Version: a04142015

# This is the R script of SINCERA analytic pipeline for analysis of single-cell RNA-seq data from heterogeneous cell populations.  
# To use this script, you will need the R statistical computing environment (version 3.1.3 or later) and several packages freely available through Bioconductor and CRAN, including
# * Biobase <http://www.bioconductor.org/packages/release/bioc/html/Biobase.html>
# * ROCR <https://rocr.bioinf.mpi-sb.mpg.de>, <http://cran.r-project.org/web/packages/ROCR/index.html>
# * RobustRankAggreg <http://cran.r-project.org/web/packages/RobustRankAggreg/index.html>
# * G1DBN <http://cran.r-project.org/web/packages/G1DBN/index.html>
# * igraph <http://igraph.org/>, <http://cran.r-project.org/web/packages/igraph/index.html> 
# SINCREA will try to resolve dependencies automatically.

# SINCERA is in the ALPHA stage of development. Core features have been implemented. We are improving the documentation, completing features, and refining the interfaces.
# SINCERA is under heavy active development. Updates of SINCERA will be distributed primarily through the SINCERA website at: http://research.cchmc.org/pbge/sincera.html.

# ================

#######################################################
#                 Resolving Dependencies              #
#######################################################

if (!require(Biobase)) {
    source("http://bioconductor.org/biocLite.R")
    biocLite(c("Biobase"))
    library(Biobase)
}
if (!require(ROCR)) {
    install.packages("ROCR")
    library(ROCR)
}
if(!require(RobustRankAggreg)) {
	install.packages("RobustRankAggreg")
	library(RobustRankAggreg)
}
if(!require(G1DBN)) {
	install.packages("G1DBN")
	library(G1DBN)
}
if(!require(igraph)) {
	install.packages("igraph")
	library(igraph)
}

#######################################################
#                   Global Variables                  #
#######################################################

dir.delim <- "/"
TIMESTAMP.FORMAT <- '%m%d%Y_%Hh%Mm%Ss'

GENE.SYMBOL.LABEL = "SYMBOL"					    # for labeling the column encoding the gene symbol info
SAMPLE.LABEL = "SAMPLE"                        		# for labeling the column encoding the sample information
CLUSTER.LABEL = "CLUSTER"  					   		# for labeling the column encoding the results of cluster assignment
TF.LABEL = "TF"										# for labeling the column encoding the transcription factors/cofactors

TTEST.MIN.SIZE = 5                             		# in differential expression test, if both groups >= 5, used Welch's t-test; otherwise, using Wilcoxon rank sum test

COMMON.TRESHOLD=5                                   # in common gene metric, expression > COMMON.TRESHOLD will considered as expressed
COMMON.PERCENTAGE=0.8                               # in common gene metric, genes that express in >COMMON.PRECENTAGE cluster cells will be considered as a common gene shared by cluster cells.
UNIQUE.RATIO=2                                      # parameters for unique gene metric
UNIQUE.QUANTILE=0.85								# parameters for unique gene metric


EXPR.SPECIFICITY.PREFIX <- "specificity_"      		# for labeling the columns encoding the results of expression specificity calculation for each group
EXPR.ABUNDANCY.PREFIX <- "abundancy_"          		# for labeling the columns encoding the results of expression abundance calculation
PREFILTERING.PREFIX <- "use_for_analysis_"	   		# for labeling the column encoding the result of prefiltering


DIFF.EXPR.PREFIX = "test_"                     		# for labeling the columns encoding the results of differential expression test for each group
CELLTYPE.ENRICHMENT.PREFIX = "use_for_celltype_"	# for labeling the columns encoding the gene list used for cell type enrichment analysis for each group
MARKER.PREFIX = "use_as_marker_"					# for labeling the columns encoding the biomarkers for each group


COMMON.PREFIX="common_"								# for labeling the columns encoding the results of common metric for each group
UNIQUE.PREFIX="unique_"								# for labeling the columns encoding the results of unique metric for each group
TEST.STATS.METRIC.PREFIX = "test_statistic_"		# for labeling the columns encoding the results of test statistic metric for each group
SYN.SIM.PREFIX="synthetic_similiary_"				# for labeling the columns encoding the results of synthetic profile similarity metric for each group
TRAINSET.PREFIX="use_for_train_"					# for labeling the columns encoding the training instances for each group
TRAIN.CLASS.PREFIX="train_class_"					# for labeling the columns encoding the class of group specific training instances
TESTSET.PREFIX="use_for_test_"						# for labeling the columns encoding the testing instances for each group
SIGNATURE.PREFIX="sig_pred_"						# for labeling the columns encoding the results of signature prediction


TG.PREFIX="use_for_tg_" 							# for labeling the columns encoding the group specific candidate regulatory targets
TF.PREFIX="use_for_tf_"								# for labeling the columns encoding the group specific candidate transcription factors/cofactors


#######################################################
#                AUXILIARY FUNCTIONS                  #
#######################################################

# format timestamp
getTimestamp <- function(format.str=TIMESTAMP.FORMAT){
	return(format(Sys.time(),format.str))
}

# Exporting specified Expression Set components
exportES <- function(ES, prefix="", suffix="", components=c("fd", "pd", "exprs")) {
	if ("fd" %in% components) {
		fd.df <- fData(ES)
		fd.df <- cbind(RID=rownames(fd.df), fd.df)
		write.table(fd.df, file=paste(prefix, "fd", suffix, sep=""), sep="\t", col.names=TRUE, row.names=FALSE)
	}
	if("pd" %in% components) {
		pd.df <- pData(ES)
		pd.df <- cbind(RID=rownames(pd.df), pd.df)
		write.table(pd.df, file=paste(prefix, "pd", suffix, sep=""), sep="\t", col.names=TRUE, row.names=FALSE)
	}
	if("exprs" %in% components) {
		exprs.m <- as.data.frame(exprs(ES))
		exprs.m <- cbind(RID=rownames(exprs.m), exprs.m)
		write.table(exprs.m, file=paste(prefix, "exprs", suffix, sep=""), sep="\t", col.names=TRUE, row.names=FALSE)
	}
}

#######################################################
#                Preprocessing                        #
#######################################################

#' Truncate expression values out of bounds to bounds
#' 
#' @param ES
#' @param upper.bound
#' @param lower.bound
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#' ES <- exprs.truncate(ES, upper.bound=NULL, lower.bound=0.01)
#' }
exprs.truncate <- function(ES, upper.bound=NULL, lower.bound=NULL, verbose=TRUE) {

    if (!is.null(upper.bound)) {
		if (verbose) {
			cat("Sincera: converting expression >", upper.bound, "to", upper.bound, "... ")
		}
        expr <- exprs(ES) 
        expr[expr > upper.bound] <- upper.bound
        exprs(ES) <- expr
		if (verbose) {
			cat("done\n")
		}
    }
    if (!is.null(lower.bound)) {
		if (verbose) {
			cat("Sincera: converting expression <", lower.bound, "to", lower.bound, "... ")
		}
        expr <- exprs(ES) 
        expr[expr < lower.bound] <- lower.bound
        exprs(ES) <- expr
		if (verbose) {
			cat("done\n\n")
		}
    }
    return(ES)
}

#' Per group expression specificity calculation
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
exprs.specificity <- function(ES, group.by=SAMPLE.LABEL, groups=NULL, specificity.prefix=EXPR.SPECIFICITY.PREFIX) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	for (i in groups) { # per group
		i.cells <- rownames(subset(pData(ES), pData(ES)[,group.by] %in% i))
		i.specificity <- apply(exprs(ES[,i.cells]), 1, function(y) specificity.helper(y))
		i.name <- paste(specificity.prefix,i,sep="")
		fData(ES)[,i.name] <- i.specificity
	}
	return(ES)
}
specificity.helper <- function(x) {
	x <- as.numeric(x)
	if (!any(x!=0)) { # if all zeros, return zero
		return(0)
	}
	specificity <- 0
	ma <- max(x)
	x <- x/ma
	x <- 1-x
	specificity <- sum(x)/(length(x)-1)
	return(specificity)
}


#' Per group expression abundancy calculation
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
exprs.abundancy <- function(ES, group.by=SAMPLE.LABEL, groups=NULL, abundancy.threshold=5, abundancy.prefix=EXPR.ABUNDANCY.PREFIX) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	for (i in groups) {
		i.cells <- rownames(subset(pData(ES), pData(ES)[,group.by] %in% i))
		i.abundancy <- apply(exprs(ES[,i.cells]), 1, function(y) abundancy.helper(y, abundancy.threshold))
		i.name <- paste(abundancy.prefix,i,sep="")
		fData(ES)[,i.name] <- i.abundancy
	}
	return(ES)
}
abundancy.helper <- function(x, threshold=5) {
	x <- as.numeric(x)
	return(length(which(x>=threshold)))
}

#' Gene prefiltering based on expression specificity and abundancy
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
prefiltering <- function(ES, 
						group.by=SAMPLE.LABEL,
						groups = NULL,
						specificity.threshold=0.7, 
						abundancy.threshold=5, 
						abundancy.count=2,
						do.batch.analysis = FALSE,
						specificity.prefix = EXPR.SPECIFICITY.PREFIX,
						abundancy.prefix = EXPR.ABUNDANCY.PREFIX,
						prefiltering.prefix=PREFILTERING.PREFIX,
						verbose=T, 
						export=T, export.components="fd") 
{
	if (verbose) {
		cat("Sincera: prefiltering start ... \n")
	}
	wd <- paste(getwd(), dir.delim, "sincera.prefiltering.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)
	
	if (verbose) {
		cat("\tMeasuring per-group specificity\n")
	}
	ES <- exprs.specificity(ES, group.by=group.by, groups=groups, specificity.prefix = EXPR.SPECIFICITY.PREFIX)
	
	if (verbose) {
		cat("\tMeasuring per-group abundancy\n")
	}
	ES <- exprs.abundancy(ES, group.by=group.by, groups=groups, abundancy.threshold=abundancy.threshold, abundancy.prefix=EXPR.ABUNDANCY.PREFIX)
	
	groups <- sort(unique(pData(ES)[,group.by]))
	s_cols <- paste(specificity.prefix, groups, sep="")
	a_cols <- paste(abundancy.prefix, groups, sep="")
	s_cols <- which(colnames(fData(ES)) %in% s_cols)
	a_cols <- which(colnames(fData(ES)) %in% a_cols)

	fData(ES)[, prefiltering.prefix] <- apply(fData(ES), 1, 
				function(y) prefiltering.helper(y, s_cols, a_cols, s_t=specificity.threshold, a_c=abundancy.count))
				
	if (do.batch.analysis == TRUE) {
	
		cat("\tGenerating Q-Q, MA, and ISCA figures before and after pre-filtering...")
	
		sample.pairs <- combn(groups, 2)
		
		for (i in 1:dim(sample.pairs)[2]) {
	
			s1.cols <- which(pData(ES)[,group.by] == sample.pairs[1,i])
			s2.cols <- which(pData(ES)[,group.by] == sample.pairs[2,i])
			
			s1.label <- as.character(sample.pairs[1,i])
			s2.label <- as.character(sample.pairs[2,i])
			
			filename=paste(wd,"QQ.", s1.label, "vs", s2.label,".before.tif", sep="")
			batch.analysis.QQ(exprs(ES), s1.cols, s2.cols, s1.label, s2.label, fig.main=paste("Q-Q of ", s1.label, " vs. ", s2.label, " (before)", sep=""), fig.filename=filename) 
			filename=paste(wd,"QQ.", s1.label, "vs", s2.label,".after.tif", sep="")
			batch.analysis.QQ(exprs(ES)[which(fData(ES)[,prefiltering.prefix]==1),], s1.cols, s2.cols, s1.label, s2.label, fig.main=paste("Q-Q of ", s1.label, " vs. ", s2.label, " (after)", sep=""), fig.filename=filename) 

			filename=paste(wd,"MA.", s1.label, "vs", s2.label,".before.tif", sep="")
			batch.analysis.MA(exprs(ES), s1.cols, s2.cols, fig.main=paste("MA of ", s1.label, " vs. ", s2.label, " (before)", sep=""), fig.filename=filename)  
			filename=paste(wd,"MA.", s1.label, "vs", s2.label,".after.tif", sep="")
			batch.analysis.MA(exprs(ES)[which(fData(ES)[,prefiltering.prefix]==1),], s1.cols, s2.cols, fig.main=paste("MA of ", s1.label, " vs. ", s2.label, " (after)", sep=""), fig.filename=filename)  
		
			file.prefix=paste(wd,"ISCA.before", sep="")
			batch.analysis.corr.dist(exprs(ES), s1.cols, s2.cols, s1.label, s2.label, file.prefix=file.prefix)
			file.prefix=paste(wd,"ISCA.after", sep="")
			batch.analysis.corr.dist(exprs(ES)[which(fData(ES)[,prefiltering.prefix]==1),], s1.cols, s2.cols, s1.label, s2.label, file.prefix=file.prefix)
		}
		
		cat(" done\n")
	}
	
	if (verbose) {
		cat("\t", length(which(fData(ES)[,prefiltering.prefix]==1)), " profiles passed both filters\n", sep="")
	}

	if (export) {
		if (verbose) {
			cat("\tExporting results to", wd, "\n")
		}
		exportES(ES, prefix=wd, suffix=".prefiltering.txt", components=export.components)
	}
	
	if (verbose) {
		cat("Sincera: prefiltering completed\n\n")
	}
	return(ES)
}	
prefiltering.helper <- function(x, s_cols, a_cols, s_t=0.7, a_c=2) {
	section <- 1
	if (any(as.numeric(x[s_cols])<s_t) | any(as.numeric(x[a_cols])<a_c)) {
		section <- 0
	}
	return(section)
}


#' Per group zscore normalization of gene expression profiles
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
normalization.zscore <- function(ES, group.by=SAMPLE.LABEL, groups=NULL, verbose=TRUE) {
	if (verbose) {
		cat("Sincera: row-by-row per-group zscore normalization ...")
	}
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	for (i in groups) {
		i.cells <- rownames(subset(pData(ES), pData(ES)[,group.by] %in% i))
		i.fpkm <- exprs(ES)[,i.cells]
		i.fpkm <- apply(i.fpkm, 1, function(y) z.transform.helper(y))
		i.fpkm <- t(i.fpkm)
		exprs(ES)[,i.cells] <- i.fpkm
	}
	if (verbose) {
		cat("done\n\n")
	}
	return(ES)
}
z.transform.helper <- function(x) {
	x <- as.numeric(x)
	x.mu <- mean(x)
	x.sd <- sd(x)
	if (x.sd == 0) {
		x <- rep(0, length(x))
	} else {
		x <- (x-x.mu)/x.sd
	}
	return(x)
}

#' Trimmed mean normalization of cellular expression profiles
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
normalization.trimmed.mean <- function(ES, group.by=SAMPLE.LABEL, groups=NULL, trim=0.2, do.scaling=TRUE) {
	x <- exprs(ES)
	trimmed.means <- apply(exprs(ES), 2, mean, trim=trim)
	exprs(ES) <- exprs(ES)/trimmed.means
	if (do.scaling==TRUE) {
		mean.trimmed.means <- mean(trimmed.means)
		exprs(ES) <- exprs(ES)*mean.trimmed.means
	}
	return(ES)
}

#' Q-Q plot of two samples
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
batch.analysis.QQ <- function(x, s1.cols, s2.cols, s1.label="s1", s2.label="S2", do.log=TRUE, fig.filename, fig.main, fig.width=1150, fig.height=1200, fig.res=300, fig.col="red") {
		
	q1 <- apply(x[,s1.cols],1,mean)
	q2 <- apply(x[,s2.cols],1,mean)
	
	if (do.log==TRUE) {
		if (any(q1==0) || any(q2==0)) {
			q1 <- q1 + 1
			q2 <- q2 + 1
		}
		q1 <- log(q1,2)
		q2 <- log(q2,2)
	}
	
	tiff(file=fig.filename, width = fig.width, height = fig.height, units = "px", res = fig.res, compression="lzw", bg = "white")
					
	mq=qqplot(q1, q2, main=fig.main, xlab=paste( s1.label, " Quantiles", sep=""), ylab=paste( s2.label, " Quantiles", sep=""), col=fig.col)
	segments(min(c(q1,q2)),min(c(q1,q2)),max(c(q1,q2)), max(c(q1,q2)), col="black")
	box(lwd=5)
	
	dev.off()
}

#' MA plot of two samples
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
batch.analysis.MA <- function(x, s1.cols, s2.cols, fig.filename, fig.main, fig.width=1150, fig.height=1200, fig.res=300, fig.col="green") {
	
	A <- 0.5*(log(apply(x[,s1.cols],1,mean),2)+log(apply(x[,s2.cols],1,mean),2))
	M <- log(apply(x[,s1.cols],1,mean),2)-log(apply(x[,s2.cols],1,mean),2)
	
	x.min <- floor(min(A))
	x.max <- ceiling(max(A))
	y.min <- floor(min(M))
	y.max <- ceiling(max(M))
	
	tiff(file=fig.filename, width = fig.width, height = fig.height, units = "px", res = fig.res, compression="lzw", bg = "white")
	
	ma <- plot(A, M, col=fig.col, main=fig.main, xlim=c(x.min,x.max), ylim=c(y.min, y.max), xlab="A", ylab="M", pch=19, cex.lab=1, cex=0.5)
	segments(x.min,0,x.max,0, col = "black", lwd=1)
	box(lwd=2)
	
	dev.off()

}

#' Analysis of inter-sample cell correlation and distance
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
batch.analysis.corr.dist<- function(x, s1.cols, s2.cols, s1.label, s2.label, file.prefix="") {
	
	cor.mat <- cor(x)
	
	s1.df <- data.frame(intra.max.cor=rep(-1, length(s1.cols)), inter.max.cor=rep(-1, length(s1.cols)), inter.cor=NA, inter.distance=NA)

	for (i in 1:length(s1.cols)) {
		s1.df$intra.max.cor[i] <- max(cor.mat[s1.cols[i], s1.cols[-i]])
		s1.df$inter.max.cor[i] <- max(cor.mat[s1.cols[i], s2.cols])
		s1.df$inter.cor[i] <- s1.df$inter.max.cor[i]
		s1.df$inter.distance[i] <- s1.df$intra.max.cor[i] - s1.df$inter.max.cor[i]
	}
	
	s1.df <- cbind(CELL=colnames(x)[s1.cols], s1.df)
	
	s2.df <- data.frame(intra.max.cor=rep(-1, length(s2.cols)), inter.max.cor=rep(-1, length(s2.cols)), inter.cor=NA, inter.distance=NA)

	for (i in 1:length(s2.cols)) {
		s2.df$intra.max.cor[i] <- max(cor.mat[s2.cols[i], s2.cols[-i]])
		s2.df$inter.max.cor[i] <- max(cor.mat[s2.cols[i], s1.cols])
		s2.df$inter.cor[i] <- s2.df$inter.max.cor[i]
		s2.df$inter.distance[i] <- s2.df$intra.max.cor[i] - s2.df$inter.max.cor[i]
	}
	
	s2.df <- cbind(CELL=colnames(x)[s2.cols], s2.df)
	
	write.table(s1.df, file=paste(file.prefix, "cor.dist.", s1.label, ".txt", sep=""), sep="\t", col.names=T, row.names=F)
	write.table(s2.df, file=paste(file.prefix, "cor.dist.", s2.label, ".txt", sep=""), sep="\t", col.names=T, row.names=F)
	
	tiff(file=paste(file.prefix, ".correlation.tif", sep=""), width = 1150, height =1200, units = "px", res = 300, compression="lzw", bg = "white")
	
	means <- c(mean(s1.df$inter.cor), mean(s2.df$inter.cor))
	y.min <- floor(min(means))
	
	mp <- barplot(means, axes=FALSE, axisnames=FALSE, ylim=c(y.min, 1), col=c("red", "blue"), main="Correlation", xlab="SAMPLE", ylab="CORRELATION")

	axis(1, labels=c("S1", "S2"), at = mp)

	axis(2, at=seq(y.min , 1, by=0.1))

	dev.off()
	
	
	tiff(file=paste(file.prefix, ".distance.tif", sep=""), width = 1150, height =1200, units = "px", res = 300, compression="lzw", bg = "white")
	
	means <- c(mean(s1.df$inter.distance), mean(s2.df$inter.distance))
	y.min <- floor(min(means))
	mp <- barplot(means, axes=FALSE, axisnames=FALSE, ylim=c(y.min, 1), col=c("red", "blue"), main="Distance", xlab="SAMPLE", ylab="CORRELATION")

	axis(1, labels=c("S1", "S2"), at = mp)

	axis(2, at=seq(y.min , 1, by=0.1))

	dev.off()
}


#######################################################
#                  Clustering                         #
#######################################################

#' Cluster Assignment
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
cluster.assignment <- function(ES, cluster.label=CLUSTER.LABEL, num.singleton=0, h=NULL, do.shift=TRUE, verbose=TRUE, export=TRUE, export.components="pd") {

	if (verbose) {
		cat("Sincera: cluster assignment start ... \n")
	}
	wd <- paste(getwd(), dir.delim, "sincera.cluster.assignment.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)

	fpkm <- exprs(ES)
	if (do.shift) {
		fpkm <- apply(fpkm, 2, function(y) y-mean(y))
	}
	dd <- as.dist((1-cor(fpkm))/2)
	hc <- hclust(dd, method="average")
	
	if (is.null(h)) {
		kk <- 0
		for (i in 2:dim(pData(ESz))[1]) {
			clusters <- cutree(hc, k=i)
			clustersizes <- as.data.frame(table(clusters))
			singleton.clusters <- which(clustersizes$Freq < 2)
			if (length(singleton.clusters) <= num.singleton) {
				kk <- i
			} else {
				break;
			}
		}
		clusters <- cutree(hc, k=kk)
	} else {
		clusters <- cutree(hc, h=h)
		kk <- length(unique(clusters))
	}
	
	if (verbose) {
		cat("\tCells are divided into", kk, "clusters\n")
		cat("\tCluster membership is encoded in pData(ES)$", cluster.label, "\n",sep="")
	}
	
	clusters <- data.frame(clusters)
	y <- data.frame(CELL=rownames(clusters), CLUSTER=clusters$clusters)
	
	pData(ES)[,cluster.label] <- y[,cluster.label]
	
	if (export) {
		if (verbose) {
			cat("\tExporting results to", wd, "\n")
		}
		exportES(ES, prefix=wd, suffix=".clusterassignment.txt", components=export.components)
	}
	
	if (verbose) {
		cat("Sincera: cluster assignment completed\n\n")
	}
	return(ES)
}

#' Ordering rows or columns according to a specified order
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
cluster.ordering <- function(ES, col.order=NULL, row.order=NULL, verbose=TRUE) {
	
	if (!is.null(col.order)) {
		if (verbose) {
			cat("Sincera: ordering columns... ")
		}
		exprs.m <- exprs(ES)[,col.order]
		cells <- pData(ES)[col.order,]
		exprs(ES) <- exprs.m
		pData(ES) <- cells
		if (verbose) {
			cat("done\n")
		}
	}
	if (!is.null(row.order)) {
		if (verbose) {
			cat("Sincera: ordering rows... ")
		}
		exprs.m <- exprs(ES)[row.order,]
		genes <- fData(ES)[row.order,]
		exprs(ES) <- exprs.m
		fData(ES) <- genes
		if (verbose) {
			cat("done\n")
		}
	}
	return(ES)
}

#' Permutation Analysis for determining significance of cluster assignments
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
cluster.permutation.analysis <- function(ES, group.by=CLUSTER.LABEL, n=20, method=c("pearson", "euclidean"), log.base=2, verbose=TRUE) {
	
	cat("Sincera: permutation analysis of cluster assignment\n")
	cat("\tgenerating", n, "random assignments for obtaining a background distribution\n")
	
	# order expression profile
	cell_order <- rownames(pData(ES))[order(pData(ES)[,group.by])]
	ES <- cluster.ordering(ES, col.order=cell_order, row.order=NULL, verbose=FALSE)
	
	# cluster size info
	cs <- as.data.frame(table(pData(ES)[, group.by]))
	colnames(cs) <- c(group.by, "SIZE")
	
	# vector to store quality scores 
	qs <- rep(NA, n+1)

	# observed assignment
	oa <- 1:dim(pData(ES))[1]
	
	# quality of observed assignment
	qs[1] <- pa.helper(ES, group.by=group.by, log.base=log.base, cs, oa,  method=method)
	
	# generate random assignments and evaluate their quality score
	for (i in 2:(n+1)) {
		# obtains a random permutation
		ra <- sample(oa, replace=FALSE)
		qs[i] <- pa.helper(ES, group.by=group.by, log.base=log.base, cs, ra,method=method)
	}
	
	qs.sd <- sd(qs[-1])
	qs.mu <- mean(qs[-1])

	cat("\tThe quality scores of random assignments has a mean=", qs.mu, " and standard deviation=", qs.sd, "\n", sep="")
	cat("\tThe quality score of the input cluster assignment is", qs[1],"\n")
	
	# using approximated normal distribution to compute p-value
	p.value <- NA
	if (qs[1]>qs.mu) {
		p.value <- 1-pnorm(qs[1], mean=qs.mu, sd=qs.sd)
	} else {
		p.value <- pnorm(qs[1], mean=qs.mu, sd=qs.sd)
	}
	cat("\tThe p-value of the quality of the input cluster assignment is ", p.value, "\n")
	cat("sincera: permutation analysis completed\n\n")
}
pa.helper <- function(ES, group.by=CLUSTER.LABEL, log.base=2, cs, a, method=c("pearson", "euclidean")) {
	s <- 0
	idx <- 1
	x.mu <- matrix(0, nrow=dim(exprs(ES))[1], ncol=length(cs[, group.by]))
	# obtain cluster centroids
	for (i in 1:length(cs[, group.by])) {
		i.x.mu <- apply(exprs(ES)[, a[idx:(idx+cs$SIZE[i]-1)]], 1, mean)
		i.s <- sum(apply(exprs(ES)[, a[idx:(idx+cs$SIZE[i]-1)]], 2, function(z) distance.helper(mu=as.numeric(i.x.mu), y=as.numeric(z), log.base=log.base, method=method)))
		s <- s + i.s
		idx <- idx+cs$SIZE[i]
	}

	return(s)
}
distance.helper <- function(mu, y, log.base=2, method=c("pearson", "euclidean")) {
	d <- NULL
	if (log.base > 1) {
		mu <- log(mu, log.base)
		y <- log(y, log.base)
	}
	if (method == "pearson") {
		d <- (1-cor(mu, y))/2
	} else if (method == "euclidean") {
		d <- sqrt(sum((y-mu)^2))
	}
	return(d)
}


#######################################################
#              Differential Expression                #
#######################################################

#' Finding group specific differentially expressed genes
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
diff.expr.test <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, diffexpr.prefix=DIFF.EXPR.PREFIX, ttest.min.size=TTEST.MIN.SIZE, do.fdr=FALSE, verbose=T, export=T, export.components="fd") {
	if (verbose) {
		cat("Sincera: diffexpr test ... \n")
	}
	wd <- paste(getwd(), dir.delim, "sincera.diffexpr.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)
	ES <- two.group.test(ES, group.by=group.by, groups=groups, diffexpr.prefix=diffexpr.prefix, ttest.min.size=ttest.min.size, do.fdr=FALSE)
	if (export) {
		if (verbose) {
			cat("\tExporting results to", wd, "\n")
		}
		exportES(ES, prefix=wd, suffix=".diffexpr.txt", components=export.components)
	}
	
	
	if (verbose) {
		cat("Sincera: diffexpr test completed\n\n")
	}
	return(ES)
}

# two-group statistical test
two.group.test <- function(ES, group.by="CLUSTER", groups=NULL, diffexpr.prefix=DIFF.EXPR.PREFIX, ttest.min.size=TTEST.MIN.SIZE, do.fdr=FALSE) {
  if (is.null(groups)) {
	groups <- sort(unique(pData(ES)[,group.by]))
  }
  cells <- rownames(pData(ES))
  for (i in groups) {
	cat("\tDifferential expression testing for group", i, "using")
	i.cells <- rownames(subset(pData(ES), pData(ES)[, group.by] %in% i))
	i.cells.o <- setdiff(cells, i.cells)
	i.cells.idx <- which(colnames(exprs(ES)) %in% i.cells)
	i.cells.o.idx <- which(colnames(exprs(ES)) %in% i.cells.o)
	if ( (length(i.cells.idx)>ttest.min.size) & (length(i.cells.o.idx)>ttest.min.size) ) {
		cat(" Welch's t-test ...")
	} else {
		cat(" Wilcoxon rank sum test ...")
	}
	i.test <- apply(exprs(ES), 1, function(y) two.group.test.helper(a=y, idx.i=i.cells.idx, idx.o=i.cells.o.idx, ttest.min.size=ttest.min.size))
	i.name <- paste(diffexpr.prefix,i,sep="")
	fData(ES)[,i.name] <- i.test
	
	if (do.fdr == TRUE) {
		i.fdr.name <- paste(diffexpr.prefix, "fdr_", i, sep="")
		fData(ES)[,i.fdr.name] <- p.adjust(i.test, method="fdr")
	}
	cat(" done\n")
  }
  return(ES)
} 
two.group.test.helper <- function(a, idx.i, idx.o, ttest.min.size=TTEST.MIN.SIZE) {
	a <- as.numeric(a)
	a_p <- 1
	if (sd(a[idx.i]) == 0 && sd(a[idx.o]) == 0 ) {
		a_p <- 1
	} else {
		if ( (length(idx.i)>ttest.min.size) & (length(idx.o)>ttest.min.size) ) {
			a_p <- t.test(a[idx.i], a[idx.o], alternative="greater", paired = FALSE, var.equal = FALSE)$p.value
		} else {
			a_p <- wilcox.test(a[idx.i], a[idx.o], alternative="greater")$p.value
		}
	}
	return(a_p)
}

#######################################################
#          Cell Type Enrichment Analysis              #
#######################################################

#' Cell type enrichement using differentially expressed genes
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' } 
celltype.enrichment <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, celltype.enrichment.prefix=CELLTYPE.ENRICHMENT.PREFIX, verbose=TRUE) {

	if (verbose) {
		cat("Sincera: cell type enrichment analysis ... \n")
	}
	wd <- paste(getwd(), dir.delim, "sincera.celltype.enrichment.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)

	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	
	x <- celltype.gene.association
	types.count <- celltype.genome.count
	
	for (i in groups) {
		if (verbose) {
			cat("\tEnriching cell types for group", i, "...")
		}
		i.de.name <- paste(celltype.enrichment.prefix, i, sep="")
		i.de <- rownames(fData(ES))[which(fData(ES)[, i.de.name]==1)]

		i.x.idx <- which(x$GENE.ID %in% i.de)
		
		if (length(i.x.idx) > 0 ) {
			i.x <- x[i.x.idx,]
			i.types <- unique(i.x$ANNOTATION.ID)
			i.types.count <- data.frame(TYPE=i.types, CL.1=NA, CL.0=NA, ALL.1=NA, ALL.0=NA, Fisher.PV=NA)
			h <- 1
			
			for (j in 1:dim(i.types.count)[1]) {
				j.type <- i.types.count$TYPE[j]
				i.j.idx <- which(i.x$ANNOTATION.ID %in% j.type)
				i.j.de <- unique(i.x$GENE.ID[i.j.idx])
				j.idx <- which(types.count$TYPE %in% j.type)
				i.types.count$CL.1[j] <- length(i.j.de)
				i.types.count$ALL.1[j] <- types.count$COUNT[j.idx]
				i.types.count$CL.0[j] <- length(i.de)-length(i.j.idx)
				i.types.count$ALL.0[j] <- length(unique(x$GENE.ID)) - types.count$COUNT[j.idx]
				f <- fisher.test(matrix(c(i.types.count$CL.1[j],i.types.count$CL.0[j],i.types.count$ALL.1[j],i.types.count$ALL.0[j]), nrow=2), alternative="greater")
				i.types.count$Fisher.PV[j] <- f$p.value
				h <- h+1
			}
		}
		if (verbose) {
			cat("done\n")
		}
		if (verbose) {
			cat("\tExporting enrichment results of group", i, "to", wd,"... ")
		}
		write.table(i.types.count, file=paste(wd, groups[i], "-celltype-enrichment.txt", sep=""), sep="\t", col.name=T, row.name=F)
		if (verbose) {
			cat("done\n")
		}
	}
	
	if (verbose) {
		cat("Sincera: cell type enrichment completed\n\n")
	}
}


#############################################################
#  Multi-marker Rank Aggregation based Cell Type Validation #
#############################################################

#' Cell type validation using the expression patterns of known biomarkers
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' } 
celltype.validation <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, marker.prefix=MARKER.PREFIX, threshold=0, verbose=TRUE) {

	if (verbose) {
		cat("Sincera: cell type validation ... \n")
	}
	wd <- paste(getwd(), dir.delim, "sincera.celltype.validation", getTimestamp(), dir.delim, sep="")
	dir.create(wd)

	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	
	cells <- rownames(pData(ES))
	types <- groups
	
	type_validation <- data.frame(Name=NULL, Score=NULL, PREDICTION=NULL, GD=NULL, TYPE=NULL)

	for (i in 1:length(types)) {
	
		i.marker.name <- paste(marker.prefix, types[i], sep="")
		i.x <- exprs(ES)[which(fData(ES)[,i.marker.name] == 1),]
		i.list <- list()
		
		for (j in 1:dim(i.x)[1]) {
			i.j.cells <- i.x[j,]
			i.j.cells <- i.j.cells[i.j.cells > threshold]
			if(length(i.j.cells) > 0) {
				i.j.cells <- sort(i.j.cells, decreasing=T)
				i.list[[as.character(rownames(i.x)[j])]] <- names(i.j.cells)
			}
		}
	
		i.cells <- aggregateRanks(glist = i.list, N = length(cells))
		
		i.cells.not.idx <- which(!(cells %in% i.cells$Name))
		if (length(i.cells.not.idx) > 0) {
			i.cells.not <- cells[i.cells.not.idx]
			i.cells.not.df <- data.frame(Name=i.cells.not, Score=1)
			rownames(i.cells.not.df) <- i.cells.not
			i.cells <- rbind(i.cells, i.cells.not.df)
		}
		
		i.cells$PREDICTION <- -log(i.cells$Score)
		i.cells$PREDICTION <- i.cells$PREDICTION/max(i.cells$PREDICTION)
		i.cells$GD <- 0
		
		i.cells.gold_standard <- rownames(pData(ES))[which(pData(ES)[,group.by] %in% types[i])]
		i.p.idx <- which(i.cells$Name %in% i.cells.gold_standard)

		if (length(i.p.idx) > 0) {
			i.cells$GD[i.p.idx] <- 1
		}
		i.pred <- prediction(i.cells$PREDICTION, i.cells$GD)
		i.perf <- performance(i.pred,"auc")
		i.auc <- i.perf@y.values[[1]]
		i.perf <- performance(i.pred, "tpr", "fpr")
		
		tiff(file=paste(wd, "celltype.validation.", i, ".tiff",sep=""), width = 1100, height =1200, units = "px", res = 300, compression="lzw", bg = "white")
		
		plot(i.perf, lwd=6, main=paste("ROC of ", as.character(types[i]), " (AUC:", round(i.auc,digit=6), ")", sep=""), cex.lab=1.2, colorize=T)
		axis(side=2, at=seq(0,1,0.1), font=2, lwd=3) 
		axis(side=1, at=seq(0,1,0.1), font=2, lwd=3) 
		box(lwd=3)
		
		dev.off()
			
		i.cells$TYPE <- as.character(types[i])
		type_validation <- rbind(type_validation, i.cells)
	}

	
	if (verbose) {
			cat("\tExporting biomarker based cell type validation results to", wd)
	}
	write.table(type_validation, file=paste(wd, "celltype.validation.txt", sep=""), sep="\t", row.names=F, col.names=T)
	if (verbose) {
			cat("  done\n")
	}
	
	if (verbose) {
		cat("Sincera: cell type validation completed \n\n")
	}
}	


#######################################################
#                 Signature Prediction                #
#######################################################

#' Predicting group specific signature genes
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
signature.analysis <- function(ES, group.by = CLUSTER.LABEL, groups=c(1,2,3,5,7,8,9),
										# training set
										trainset.prefix = TRAINSET.PREFIX,
										train.class.prefix = TRAIN.CLASS.PREFIX,
										marker.prefix=MARKER.PREFIX,
										do.cc.validation = TRUE,
										# testing set
										testset.prefix=TESTSET.PREFIX,
										# common gene metric
										common.prefix = COMMON.PREFIX,
										common.threshold=COMMON.TRESHOLD, common.percentage=COMMON.PERCENTAGE,
										# unique gene metric
										unique.prefix = UNIQUE.PREFIX,
										unique.ratio=UNIQUE.RATIO, unique.quantile=UNIQUE.QUANTILE,
										# test statistic metric
										test.statistic.metric.prefix = TEST.STATS.METRIC.PREFIX, log.base=2,
										diff.expr.prefix=DIFF.EXPR.PREFIX,
										diff.expr.threshold = 0.05,
										# synthetic profile similarity metric 
										syn.sim.prefix = SYN.SIM.PREFIX,
										signature.prefix = SIGNATURE.PREFIX,
										gene.symbol.label = GENE.SYMBOL.LABEL,
										verbose=TRUE,
										export=TRUE, export.components="fd" ) {
	
	if (verbose) {
		cat("Sincera: gene signature prediction ... \n")
	}
	wd <- paste(getwd(), dir.delim, "sincera.signature.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)
	
	ES <- exprs.test.statistic.metric(ES, group.by=group.by, groups=groups, log.base=log.base, test.statistic.metric.prefix=test.statistic.metric.prefix, diff.expr.prefix=diff.expr.prefix) 
	
	ES <- exprs.common(ES, group.by=group.by, groups=groups, common.threshold=common.threshold, common.percentage=common.percentage, common.prefix=common.prefix)
	
	ES <- exprs.unique(ES, group.by=group.by, groups=groups, unique.ratio=unique.ratio, unique.quantile=unique.quantile, unique.prefix=unique.prefix)

	ES <- exprs.synthetic.similarity(ES, group.by=group.by, groups=groups, syn.sim.prefix=syn.sim.prefix)

	ES <- sigpred.training.sets(ES, group.by=group.by, groups=groups, 
									trainset.prefix=trainset.prefix,
									train.class.prefix=train.class.prefix,									
									test.statistic.metric.prefix = test.statistic.metric.prefix,
									common.prefix = common.prefix,
									unique.prefix = unique.prefix,
									syn.sim.prefix = syn.sim.prefix,
									marker.prefix = marker.prefix,
									verbose=T,
									export=T
									)
	
	ES <- sigpred.testsets(ES, group.by=group.by, groups=groups, testset.prefix=testset.prefix, diff.expr.prefix=diff.expr.prefix, diff.expr.threshold=diff.expr.threshold)
		
	ES <- signature.prediction(ES, group.by=group.by, groups=groups, 
								trainset.prefix=trainset.prefix,
								train.class.prefix=train.class.prefix,
								do.cc.validation=do.cc.validation, 
								testset.prefix=testset.prefix,								
								test.statistic.metric.prefix = test.statistic.metric.prefix,
								common.prefix = common.prefix,
								unique.prefix = unique.prefix,
								syn.sim.prefix = syn.sim.prefix,
								signature.prefix = signature.prefix,
								gene.symbol.label = gene.symbol.label,
								verbose=T,
								export=T,
								wd = wd
								) 
	if (export) {
		if (verbose) {
			cat("\tExporting results to", wd, "\n")
		}
		exportES(ES, prefix=wd, suffix=".sigpred.txt", components=export.components)
	}
	
	if (verbose) {
		cat("Sincera: gene signature prediction completed\n\n")
	}
	return(ES)

}



#' Identifying group specific common genes
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
exprs.common <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, common.threshold=5, common.percentage=0.8, common.prefix=COMMON.PREFIX) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	for (i in groups) {
		i.cells <- rownames(subset(pData(ES), pData(ES)[, group.by] %in% i))
		i.cells.idx <- which(colnames(exprs(ES)) %in% i.cells)
		i.common <- apply(exprs(ES), 1, function(y) common.helper(y, i.cells.idx, common.threshold, common.percentage))
		i.name <- paste(common.prefix, i, sep="")
		fData(ES)[,i.name]<-i.common
	}
	return(ES)
}
common.helper <- function(a, idx.i, common.threshold=5, common.percentage=0.8) {
	a <- as.numeric(a)
	a_common_threshold <- ceiling(length(idx.i)*common.percentage)
	a_common <- 0
	if (length(which(a[idx.i] >= common.threshold)) >= a_common_threshold) {
		a_common <- 1
	}
	return(a_common)
}


#' Identifying group specific unique genes
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
exprs.unique <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, unique.ratio=2, unique.quantile=0.85, unique.prefix=UNIQUE.PREFIX) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	cells <- rownames(pData(ES))
	for (i in groups) {
		i.cells <- rownames(subset(pData(ES), pData(ES)[, group.by] %in% i))
		i.cells.o <- setdiff(cells, i.cells)
		i.cells.idx <- which(colnames(exprs(ES)) %in% i.cells)
		i.cells.o.idx <- which(colnames(exprs(ES)) %in% i.cells.o)
		i.unique <- apply(exprs(ES), 1, function(y) unique.helper(y, i.cells.idx, i.cells.o.idx, unique.ratio, unique.quantile))
		i.name <- paste(unique.prefix, i, sep="")
		fData(ES)[,i.name]<-i.unique
	}
	return(ES)
}
unique.helper <- function(a, idx.i, idx.o, unique.ratio=2, unique.quantile=0.85) {
	a <- as.numeric(a)
	a_mu <- mean(a[idx.i])
	a_qn <- quantile(a[idx.o], unique.quantile)[[1]]
	a_unique <- 0
	if (a_mu/a_qn >= unique.ratio) {
		a_unique <- 1
	}
	return(a_unique)
}


#' Normalized and smoothened results of group specific differential expression test
#' 
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
exprs.test.statistic.metric <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, log.base=2, test.statistic.metric.prefix=TEST.STATS.METRIC.PREFIX, diff.expr.prefix=DIFF.EXPR.PREFIX) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	#cells <- rownames(pData(ES))
	for (i in groups) {
		i.name <- paste(test.statistic.metric.prefix, i, sep="")
		i.diffexpr.test.name <- paste(diff.expr.prefix, i, sep="")
		
		fData(ES)[,i.name] <- fData(ES)[, i.diffexpr.test.name]
		fData(ES)[,i.name] <- -log(fData(ES)[,i.name], log.base)
		i.min <- min(fData(ES)[,i.name])
		i.max <- max(fData(ES)[,i.name])
		fData(ES)[,i.name] <- fData(ES)[,i.name]/i.max
	}
	return(ES)
}


#' Calculating the similarity between each gene profile and group specific synthetic reference expression profile
#' synthetic profile is created by setting fpkm(cluster cells)=1, fpkm(non cluster cells)=0
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
exprs.synthetic.similarity <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, syn.sim.prefix=SYN.SIM.PREFIX) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	cells <- rownames(pData(ES))
	for (i in groups) {
		i.cells <- rownames(subset(pData(ES), pData(ES)[, group.by] %in% i))
		i.cells.o <- setdiff(cells, i.cells)
		i.cells.idx <- which(colnames(exprs(ES)) %in% i.cells)
		i.cells.o.idx <- which(colnames(exprs(ES)) %in% i.cells.o)
		i.synthetic.profile <-  rep(0, dim(exprs(ES))[2])
		i.synthetic.profile[i.cells.idx] <- 1
		i.ss <- apply(exprs(ES), 1, function(y) (1+cor(i.synthetic.profile, as.numeric(y)))/2)
		i.name <- paste(syn.sim.prefix, i, sep="")
		fData(ES)[,i.name]<-i.ss
	}
	return(ES)
}

#' Constructing group specific training sets for signature prediction
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
sigpred.training.sets <- function(ES, group.by=CLUSTER.LABEL, groups=NULL, 
									trainset.prefix=TRAINSET.PREFIX,
									train.class.prefix=TRAIN.CLASS.PREFIX,									
									test.statistic.metric.prefix = TEST.STATS.METRIC.PREFIX,
									common.prefix = COMMON.PREFIX,
									unique.prefix = UNIQUE.PREFIX,
									syn.sim.prefix = SYN.SIM.PREFIX,
									marker.prefix = MARKER.PREFIX,
									verbose=T,
									export=T
									) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}

	for (i in groups) {
		i.train.name <- paste(trainset.prefix, i, sep="")
		fData(ES)[,i.train.name] <- 0
		i.train.class <- paste(train.class.prefix, i, sep="")
		fData(ES)[,i.train.class] <- NA
		
		i.p.candidates <- c()
		i.n.candidates <- c()
		
		i.test.name <- paste(test.statistic.metric.prefix, i, sep="")
		i.common.name <- paste(common.prefix, i, sep="")
		i.unique.name <- paste(unique.prefix, i, sep="")
		i.ss.name <- paste(syn.sim.prefix, i, sep="")
		
		fd <- fData(ES)
		fd <- fd[order(fd[,i.test.name]),]
		
		i.marker.name <- paste(marker.prefix, i, sep="")
		#i.p.symbols <- markers$SYMBOL[which(markers$CLUSTER %in% i)]
		i.p.candidates <- rownames(fd)[which(fd[, i.marker.name] == 1)]
		
		lambda_n =length(i.p.candidates)
		
		fd <- fData(ES)
		#fd <- fd[order(fd[,i.test.name], decreasing=T),]
		fd <- fd[order(fd[,i.test.name], decreasing=F),]
		fd <- subset(fd, fd[,i.common.name]==0 & fd[,i.unique.name]==0)
		i.n.candidates <- rownames(fd)[1:lambda_n]
		
		if (lambda_n > 0) {
		
		i.p.idx <- which(rownames(fData(ES)) %in% i.p.candidates)
		i.n.idx <- which(rownames(fData(ES)) %in% i.n.candidates)
		
		fData(ES)[c(i.p.idx, i.n.idx), i.train.name] <- 1
		fData(ES)[i.p.idx, i.train.class] <- 1
		fData(ES)[i.n.idx, i.train.class] <- 0
		}
	}
	return(ES)
}


#' Constructing group specific testing sets for signature prediction
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
sigpred.testsets <- function(ES, group.by="CLUSTER",  groups=NULL, testset.prefix=TESTSET.PREFIX, diff.expr.prefix=DIFF.EXPR.PREFIX, diff.expr.threshold=0.05) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	for (i in groups) {
		i.name <- paste(testset.prefix, i, sep="")
		fData(ES)[,i.name] <- 0
		i.test.name <- paste(diff.expr.prefix, i, sep="")
		i.test.idx <- which(fData(ES)[,i.test.name] < diff.expr.threshold)
		if (length(i.test.idx) > 0) {
			fData(ES)[i.test.idx, i.name] <- 1
		}
	}
	return(ES)
}



#' logistic regression based signature prediction and cross cluster validation
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
signature.prediction <- function(ES, group.by="CLUSTER", groups=NULL, do.cc.validation=TRUE, 
								trainset.prefix=TRAINSET.PREFIX,
								train.class.prefix=TRAIN.CLASS.PREFIX,
								testset.prefix=TESTSET.PREFIX,								
								test.statistic.metric.prefix = TEST.STATS.METRIC.PREFIX,
								common.prefix = COMMON.PREFIX,
								unique.prefix = UNIQUE.PREFIX,
								syn.sim.prefix = SYN.SIM.PREFIX,
								signature.prefix = SIGNATURE.PREFIX,
								gene.symbol.label = GENE.SYMBOL.LABEL,
								verbose=T,
								export=T,
								wd = wd
								) {
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	predictions <- list()	
	for (i in groups) {
	
		i.train.name <- paste(trainset.prefix, i, sep="")
		
		i.test.name <- paste(test.statistic.metric.prefix, i, sep="")
		i.common.name <- paste(common.prefix, i, sep="")
		i.unique.name <- paste(unique.prefix, i, sep="")
		i.ss.name <- paste(syn.sim.prefix, i, sep="")
		i.train.class <- paste(train.class.prefix, i, sep="")
	
		# train
		i.train.data <- subset(fData(ES)[, c(i.common.name, i.unique.name, i.test.name, i.ss.name, i.train.class)], fData(ES)[, i.train.name] == 1)
		
		if (verbose == T) {
			cat("\tConstructing model for cluster", i,":")
		}
		
		formula_for_train <- paste(i.train.class, "~", i.test.name, "+", i.ss.name, sep="")

		i.train.p <- i.train.data[which(i.train.data[,i.train.class]==1),]
		i.train.p.sd <- apply(as.matrix(i.train.p), 2, sd)
		
		# removing dominant factors
		
		factors <- c(i.common.name, i.unique.name, i.test.name, i.ss.name)
		
		if (i.train.p.sd[i.common.name] == 0) {
			cat("ignore common gene metric ")
		} else {
			formula_for_train <- paste(formula_for_train, "+", i.common.name, sep="")
		}
		
		if (i.train.p.sd[i.unique.name] == 0) {
			cat(" ignore unique gene metric ")
		} else {
			formula_for_train <- paste(formula_for_train, "+", i.unique.name, sep="")
		}
		
		i.glm <- glm(as.formula(formula_for_train), 
						data=i.train.data, 
						family = "binomial")
						
		cat(" done\n")

		# cross-cluster validation
		if (do.cc.validation) {
			cat("\tCross cluster validating the constructed signature prediction system of group", i," ")
			i.testing.name <- paste(testset.prefix, i, sep="")
			i.positive.idx <- which(fData(ES)[,i.testing.name] == 1)
			i.positive.data <- fData(ES)[i.positive.idx, c(gene.symbol.label, i.common.name, i.unique.name, i.test.name, i.ss.name)]
			i.positive.data$g <- 1
			#cat("C1-p: ", dim(i.positive.data)[1], " ; ", sep="")

			negative.groups <- setdiff(groups, i)
			i.n <- length(negative.groups)
				
			for (j in negative.groups) {
			
				i.j.validation.data <- i.positive.data
			
				j.testing.name <- paste(testset.prefix, j, sep="")
				j.negative.idx <- which(fData(ES)[,j.testing.name] == 1)
				j.negative.idx <- j.negative.idx[!is.na(j.negative.idx)]
				j.negative.idx <- unique(j.negative.idx)
				j.negative.data <- fData(ES)[j.negative.idx, c(gene.symbol.label, i.common.name, i.unique.name, i.test.name, i.ss.name)]
				j.negative.data$g <- 0
				
				overlapped <- intersect(rownames(i.j.validation.data), rownames(j.negative.data))
				if (length(overlapped) > 0 ) {
					i.j.validation.data <- i.j.validation.data[which(!(rownames(i.j.validation.data) %in% overlapped)),]
					
					# NOTE: commented to reduce the performance of cross-cluster validations
					# j.negative.data <- j.negative.data[which(!(rownames(j.negative.data) %in% overlapped)),]
				}
				
				#cat(j," (", dim(j.negative.data)[1], "), ", sep="")
				i.j.validation.data <- rbind(i.j.validation.data, j.negative.data)
				
				i.j.validation.data <- i.j.validation.data[which(!(rownames(i.j.validation.data) %in% rownames(i.train.data))),]
						
				i.j.validates <- predict(i.glm, i.j.validation.data)
				i.j.validates <- as.numeric(i.j.validates)
				#write.table(data.frame(GENE=rownames(i.j.validation.data), SYMBOL=i.j.validation.data$SYMBOL, P=as.numeric(i.j.validates)), file=paste(wd , "C", i, "-C", j, ".txt", sep=""), sep="\t", col.names=T, row.names=F)
								
				# standardize score
				i.j.validates <- (i.j.validates-min(i.j.validates))/(max(i.j.validates)-min(i.j.validates))
				
				i.j.p <- prediction(i.j.validates, i.j.validation.data$g)
				perf <- performance(i.j.p,"auc")
				i.j.auc <- perf@y.values[[1]]
				perf <- performance(i.j.p, "tpr", "fpr")
				
				tiff(file=paste(wd, "cross-cluster-validation-", i, "-", j, ".tiff",sep=""), width = 1100, height =1200, units = "px", res = 300, compression="lzw", bg = "white")
				
				plot(perf, lwd=7, main=paste("ROC of C", i, "-C", j, " (AUC:", round(i.j.auc,digit=6), ")", sep=""), cex.lab=1.5, colorize=T)
				axis(side=2, at=seq(0,1,0.1), font=2, lwd=3) 
				axis(side=1, at=seq(0,1,0.1), font=2, lwd=3) 
				box(lwd=3)
				
				dev.off()
			}
			cat("done\n")

		}

		cat("\tpredicting signature genes for group", i, " ")
		
		# prediction
		i.testing.name <- paste(testset.prefix, i, sep="")
		i.test.data <- subset(fData(ES)[, c(gene.symbol.label, i.common.name, i.unique.name, i.test.name, i.ss.name)], fData(ES)[,i.testing.name]==1) 
		i.prediction <- predict(i.glm, i.test.data)
		i.pred.name <- paste(signature.prefix, i, sep="")
		#predictions[i.pred.name] <- data.frame(GENE=i.test.data$SYMBOL, P=as.numeric(i.prediction))
		
		i.prediction <- as.numeric(i.prediction)
		i.prediction <- (i.prediction-min(i.prediction))/(max(i.prediction)-min(i.prediction))
		
		fData(ES)[,i.pred.name] <- NA
		fData(ES)[rownames(i.test.data), i.pred.name] <- i.prediction
		write.table(data.frame(GENE=rownames(i.test.data), SYMBOL=i.test.data$SYMBOL, P=as.numeric(i.prediction)), file=paste(wd , "signature_prediction_", i, ".txt", sep=""), sep="\t", col.names=T, row.names=F)
		
		cat("done\n")
	}
	
	#print(glms)
	return(ES)
}


#######################################################
#             Driving Force Analysis                  #
#######################################################

#' Inferring first order conditional dependencies between TFs and TGs
#' @param ES
#' @param verbose
#' @return
#' @details
#' @export
#' @examples
#' \dontrun{
#' load("E16.5.Rd")
#' pd <- new("AnnotatedDataFrame", data = cells)
#' fd <- new("AnnotatedDataFrame", data = genes)
#' ES <- new("ExpressionSet", exprs = as.matrix(exprs.m), phenoData = pd, featureData = fd)
#'
#' }
network.inference <-function(data, method='ls',predPosition=NULL, targetPosition=NULL, lag=0) {

  ## ===============================================
  ## INITIALIZING
  ## _______________________________________________

  data<-as.matrix(data)
  n=dim(data)[1] # nb of time points
  p=dim(data)[2] # nb of genes
  
  ## If predictor or target genes position is specified,
  ## we just work of them
  if(is.null(predPosition)){
    pred <- 1:p}
  else {
    pred <- predPosition}
  d <- length(pred)
  
  if(is.null(targetPosition)){
    tar <- 1:p}
  else {
    tar <- targetPosition}
  r <- length(tar)

  ## The score matrix
  S1ls=NULL
  S1tukey=NULL
  S1huber=NULL
  
  if('ls' %in% method){
    S1ls<-matrix(0,r,d)
  }
  if('tukey' %in% method){
    S1tukey<-matrix(0,r,d)
  }
  if('huber' %in% method){
    S1huber<-matrix(0,r,d)
  }

  ## ===============================================
  ## BUILDING THE SCORE MATRICES
  ## _______________________________________________

  ## Print the total number of vertices

  #cat("Treating", r ,"vertices:\n")
  cpt=10
 
  for (i in 1:r){
    ## Print percentage of accomplished vertices
    if ( ((i/r)*100)>=cpt ) {
      #cat(cpt, "% ",sep="")
      cpt=cpt+10
    }
  
   
    ## The regression model is
    ## Y = X
    
    ## Y is the vector containing the target genes
    ## on which the regression will be performed
    ## The time point 1 is removed
	if (1+lag >= n) {
		stop();
	}
    y<-data[(1+lag):n,tar[i]]

    for (j in  1:(d-1)){
      
      ## for all k > j
      for (k in c(1:d)[-c(1:j)]){

        ## X is a matrix with two columns containing the
        ## predicted gene on which the regression will
        ## be performed. The two columns contain
        ## respectively the data corresponding to the jth
        ## and the kth tested gene.
        ## The time n is removed
        x<-data[1:(n-lag),c(pred[j],pred[k])]
        
        ## =====================================
        ## ESTIMATION...
        ## 
        ## Three estimators are available :
        ## - Least square
        ## - Huber
        ## - Tuckey
        
        ## =====================================
        ## LEAST SQUARE ESTIMATOR
        if('ls' %in% method){
          lm.3<-lm(y~x)
          prob<-abs(summary(lm.3)$coef[,"Pr(>|t|)"])  
          ## coefficient aij(k) : aij given k 
          S1ls[i,j]<-max(prob[2],S1ls[i,j],na.rm=TRUE)
          ## coefficient aik(j) : aik given j 
          S1ls[i,k]<-max(prob[3],S1ls[i,k],na.rm=TRUE)
        }
        
        ## =====================================
        ## TUKEY'S ESTIMATOR
        if('tukey' %in% method){
          bisq.3<-rlm(y~x,method='MM')
          prob<-2*(1-pt(abs(summary(bisq.3)$coef[,"t value"]),n-4))
          ## coefficient aij(k) : aij given k 
          S1tukey[i,j]<-max(prob[2],S1tukey[i,j],na.rm=TRUE)
          ## coefficient aik(j) : aik given j 
          S1tukey[i,k]<-max(prob[3],S1tukey[i,k],na.rm=TRUE)
        }

        ## =====================================
        ## HUBER'S ESTIMATOR
        if('huber' %in% method ){
          hub.3<-rlm(y~x)
          prob<-2*(1-pt(abs(summary(hub.3)$coef[,"t value"]),n-4))
          ## coefficient aij(k) : aij given k 
          S1huber[i,j]<-max(prob[2],S1huber[i,j],na.rm=TRUE)
          ## coefficient aik(j) : aik given j 
          S1huber[i,k]<-max(prob[3],S1huber[i,k],na.rm=TRUE)
       }

         
      } # end k
    } # end j
  } # end i 
  #cat("\n")

  ## The score matrices are return
  list(S1ls=S1ls,S1tukey=S1tukey,S1huber=S1huber)
}

#' Calculate the importance of a node in a graph
# 
#' @param g an igraph object
#' @param measurement DFC, DCC, DDC, DC, CC, BC
#' @param componentType character string, either weak or strong; for directed graphs, weak implies weakly, strong strongly connected components to search. 
#' @param pathType defined the types of the paths used for measuring the distance in directed graphs. 
#' @param directed whether the graph is directed or undirected;
#' @param normalized whether to rescale the results to [0,1]
#' @return a importance matrix

getNodesImportance <- function(
	graph, nodes=NULL, measurements=c("DFC","DCC","DDC","DC","CC","BC"),
	componentType="weak", 
	pathType="all", 
	directed=FALSE, 
	normalized=TRUE) 
{	
	cat("\tCalculating importance measurements: ")
	if (length(nodes)==0) {
		nodes=V(graph)
	}
	im <- as.data.frame(matrix(0, length(nodes), length(measurements)))
	rownames(im) <- nodes$name
	colnames(im) <- measurements
	
	for (m in measurements) {
		if (m == "DFC") {
			for (i in nodes) {
				rg <- delete.vertices(graph, i)
				n <- vcount(rg)
				cls <- clusters(rg, mode=componentType) # mode: strong, weak
				im[i,m]<- cls$no/n
			}
			cat("DFC ")
		} else if (m == "DCC") {
			for (i in nodes) {
				rg <- delete.vertices(graph, i)
				n <- vcount(rg)
				cls <- clusters(rg, mode=componentType) # mode: strong, weak
				s <- 0
				for (cs in cls$csize) {
					s <- s + cs*(cs-1)
				}
				im[i,m]<- 1 - s/(n*(n-1))
			}
			cat("DCC ")
		} else if (m == "DDC") {
			for (i in nodes) {
				rg <- delete.vertices(graph, i)
				n <- vcount(rg)
				spl <- shortest.paths(rg, mode=pathType) #mode: in, out, all
				s <- 0
				for ( si in 1:dim(spl)[1]) {
					for (sj in 1:dim(spl)[2]) {
						if (si == sj) {
						} else {
							if(spl[si,sj] == Inf) {
								s <- s + (1/n)
							} else {
								s <- s + (1/spl[si,sj])
							}
						}
					}
				}
				im[i,m]<- 1 - s/(n*(n-1))	
			}
			cat("DDC ")
		} else if (m == "DC") {
			dgrs <- degree(graph,nodes,mode=pathType,loops=TRUE,normalized=normalized) # mode: in, out, all
			im[,m] <- dgrs
			cat("DC ")
		} else if (m == "CC") {
			clsns <- closeness(graph,nodes,mode=pathType,normalized=normalized) # mode: in, out, all
			im[,m] <- clsns
			cat("CC ")
		} else if (m == "BC") {
			btwns <- betweenness(graph, nodes, directed=directed, normalized=normalized) 
			im[,m] <- btwns
			cat("BC ")
		} else {
			stop(cat("Unrecognized Importance Measurements:", m))
		}
	}
	cat("\n")
	return(im)
}

#' load graph from edgelist file and extract the largest connected component
getGraph <- function(filename, threshold, directed=FALSE) {
	#cat("\tLoading network from ",filename, "\n", sep="")
	edges <- read.table(file=filename, sep="\t", header=T)
	edges <- edges[which(edges$s1<threshold),]
	g <- graph.data.frame(edges, directed=directed)
	cl <- clusters(g, mode="weak")
	g <- induced.subgraph(g, which(cl$membership == which.max(cl$csize)))
	#cat("\tLoading completes!\n")
	return(g)
}

#' interface to the driving force analysis
driving.force.analysis <- function(ES, group.by="CLUSTER", groups=NULL, unique.by=NULL, norm.method="zscore", prefix.tg=TG.PREFIX, prefix.tf=TF.PREFIX, verbose=T) {
	if (is.null(unique.by)) {
		driving.force.analysis.1(ES, group.by=group.by, groups=groups, prefix.tg=prefix.tg, prefix.tf=prefix.tf, verbose=T)
	} else {
		driving.force.analysis.2(ES, group.by=group.by, unique.by=unique.by, norm.method="zscore", groups=groups, prefix.tg=prefix.tg, prefix.tf=prefix.tf, verbose=T)
	}
}


#' construct per-group transcriptional regulatory network (implementation 2)
#' @param prefix.tg prefix for labeling the columns that specify candidate group specific regulatory targets
#' @param prefix.tf prefix for labeling the columns that specify candidate group specific transcriptional factors
driving.force.analysis.2 <- function(ES, group.by="CLUSTER", groups=NULL, unique.by="SYMBOL", norm.method="zscore", prefix.tg="use_for_tg_", prefix.tf="use_for_tf_", verbose=T) {

	# if no groups are specified, construct TRN for all groups
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	
	if (verbose) {
		cat("Sincera: driving force analysis ... \n")
	}
	
	wd <- paste(getwd(), dir.delim, "sincera.drivingforce.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)
	
	identifier <- unique.by
	
	genes <- fData(ES)
	cells <- pData(ES)
	fpkm <- as.data.frame(exprs(ES))	
	fpkm[, identifier] <- genes[,identifier]
	
	for (i in groups) {
	
		cat("\tinferring TRN of group", i, "from the expression patterns of ")

		i.tg.colname <- paste(prefix.tg, i, sep="")
		i.tf.colname <- paste(prefix.tf, i,sep="")
		
		i.tgs <- rownames(subset(genes, genes[,i.tg.colname] ==1))
		i.cells <- rownames(subset(cells, cells[,group.by] %in% i))
		i.tfs <- rownames(subset(genes, genes[,i.tf.colname]==1))
		
		
		
		i.genes <- union(i.tgs, i.tfs)
		
		i.exprs <- fpkm[i.genes,]
		
		i.exprs <- aggregate(. ~ i.exprs[,identifier], data=i.exprs, FUN=mean)
		colnames(i.exprs)[1] <- identifier
		rownames(i.exprs) <- i.exprs[,identifier]
		i.exprs <- i.exprs[-1]
		i.exprs <- i.exprs[-(dim(i.exprs)[2])]
		
		if (norm.method=="zscore") {	
			# per-sample zscore
			samples <- sort(unique(pData(ES)[,SAMPLE.LABEL]))
			for (j in samples) {
				j.cells <- rownames(subset(pData(ES), pData(ES)$SAMPLE %in% j))
				j.fpkm <- i.exprs[,j.cells]
				j.fpkm <- apply(j.fpkm, 1, function(y) z.transform.helper(y))
				j.fpkm <- t(j.fpkm)
				i.exprs[,j.cells] <- j.fpkm
			}
		} else if(norm.method=="log2") {
			i.exprs <- log(i.exprs,2)
		}
		
		i.tgs.symbol <- unique(genes[i.tgs, identifier])
		i.tfs.symbol <- unique(genes[i.tfs, identifier])
		
		cat(length(i.tgs.symbol), " TGs and ", length(i.tfs.symbol), " TFs: ", sep="") 
		
		i.tgs.idx <- which(rownames(i.exprs) %in% i.tgs.symbol)
		i.tfs.idx <- which(rownames(i.exprs) %in% i.tfs.symbol)
		
		i.exprs <- i.exprs[,i.cells]
		i.exprs <- t(i.exprs)
		
		tfs.symbols <- colnames(i.exprs)[i.tfs.idx]
	
		mat.file <- paste(wd, "AdjMat_", i, ".txt",sep="")
		write(paste("TG",paste(tfs.symbols,collapse="\t"),collapse="\t", sep="\t"),file=mat.file)
	
		cat("0% ")
		
		z <- 1
		cpt <- 10
		for (u in i.tgs.idx) {
			# progress report
			if ( ((z/length(i.tgs.idx)*100))>=cpt ) {
			  cat(cpt, "% ",sep="")
			  cpt=cpt+10
			}
			S1 <- network.inference(i.exprs, method="ls",predPosition=i.tfs.idx,target=u,lag=0)
			write(paste(colnames(i.exprs)[u],paste(S1$S1ls[,],collapse="\t"),collapse="\t",sep="\t"),file=mat.file, append=T)
			z <- z+1
		}
		cat("\n")
	
		i.s1 <- read.table(file=mat.file, sep="\t", head=T, row.name=1,check.names=FALSE, quote="")
	
		edge.file <- paste(wd, "EdgeList_", i, ".txt", sep="")
		write("TF\tTG\ts1", file=edge.file)
		for (u in 1:dim(i.s1)[2]) {
			for (v in 1:dim(i.s1)[1]) {
				u.symbol <- as.character(colnames(i.s1)[u])
				v.symbol <- as.character(rownames(i.s1)[v])
				if (u.symbol != v.symbol) {
					write(paste(u.symbol, v.symbol, i.s1[v,u], sep="\t"), file=edge.file, append=T)
				}
			}
		}
		
		g <- getGraph(edge.file, 0.05, FALSE)
		TF.importance.metrics = c("DFC","DCC", "DDC", "DC", "CC", "BC")
		im <- getNodesImportance(g,measurements=TF.importance.metrics,
			componentType="weak", 
			pathType="all", 
			directed=FALSE, 
			normalized=TRUE)
		
		
		write.table(cbind(RID=rownames(im),im), file=paste(wd, "NodeIm.in.TRN_", i,".LCC.txt", sep=""), col.names=T, row.names=F, sep="\t")
		
		im$TF <- 0
		for (j in 1:dim(im)[1]) {
			idx <- which(fData(ES)[,identifier] %in% rownames(im)[j])
			if (length(idx) > 0) {
				if (any(fData(ES)$TF[idx] == 1)) {
					im$TF[j] <- 1
				}
			}
		}
		im <- im[which(im$TF == 1),TF.importance.metrics]
		
		cat("\tfinding important TFs in the main connected component of TRN of group", i, "\n")
		
		im.r <- as.data.frame(apply(im, 2, function(y) rank(-y, ties.method="min")))
		avg.r <- apply(im.r, 1, mean)
		im.r$AVG.RANK <- as.numeric(rank(avg.r, ties.method="min"))
		
		im.r <- im.r[order(im.r$AVG.RANK),]
		
		im.r <- cbind(TF=rownames(im.r), im.r)
		
		write.table(im.r, file=paste(wd, "TF.ranks.in.TRN_", i,".LCC.txt", sep=""), col.names=T, row.names=F, sep="\t")

	}
	
	if (verbose) {
		cat("Sincera: driving force analysis completed \n\n")
	}
}

#' construct per-group transcriptional regulatory network (implementation 1)
#' @param prefix.tg prefix for labeling the columns that specify candidate group specific regulatory targets
#' @param prefix.tf prefix for labeling the columns that specify candidate group specific transcriptional factors
driving.force.analysis.1 <- function(ES, group.by="CLUSTER", groups=NULL, prefix.tg="use_for_tg_", prefix.tf="use_for_tf_", verbose=T) {

	# if no groups are specified, construct TRN for all groups
	if (is.null(groups)) {
		groups <- sort(unique(pData(ES)[,group.by]))
	}
	
	if (verbose) {
		cat("Sincera: driving force analysis ... \n")
	}
	
	wd <- paste(getwd(), dir.delim, "sincera.drivingforce.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)
	
	genes <- fData(ES)
	cells <- pData(ES)
	fpkm <- as.data.frame(exprs(ES))	
	
	for (i in groups) {
	
		cat("\tinferring TRN of group", i, "from the expression patterns of ")

		i.tg.colname <- paste(prefix.tg, i, sep="")
		i.tf.colname <- paste(prefix.tf, i,sep="")
		
		i.tgs <- rownames(subset(genes, genes[,i.tg.colname] ==1))
		i.cells <- rownames(subset(cells, cells[,group.by] %in% i))
		i.tfs <- rownames(subset(genes, genes[,i.tf.colname]==1))
		
		cat(length(i.tgs), " TGs and ", length(i.tfs), " TFs: ", sep="") 
		
		i.genes <- union(i.tgs, i.tfs)
		
		i.exprs <- fpkm[i.genes,]
		
		i.tgs.idx <- which(rownames(i.exprs) %in% i.tgs)
		i.tfs.idx <- which(rownames(i.exprs) %in% i.tfs)
		
		i.exprs <- i.exprs[,i.cells]
		i.exprs <- t(i.exprs)
		
		tfs.symbols <- colnames(i.exprs)[i.tfs.idx]
	
		mat.file <- paste(wd, "AdjMat_", i, ".txt",sep="")
		write(paste("TG",paste(tfs.symbols,collapse="\t"),collapse="\t", sep="\t"),file=mat.file)
	
		z <- 1
		cpt <- 10
		for (u in i.tgs.idx) {
			# progress report
			if ( ((z/length(i.tgs.idx)*100))>=cpt ) {
			  cat(cpt, "% ",sep="")
			  cpt=cpt+10
			}
			S1 <- network.inference(i.exprs, method="ls",predPosition=i.tfs.idx,target=u,lag=0)
			write(paste(colnames(i.exprs)[u],paste(S1$S1ls[,],collapse="\t"),collapse="\t",sep="\t"),file=mat.file, append=T)
			z <- z+1
		}
		cat("\n")
	
		i.s1 <- read.table(file=mat.file, sep="\t", head=T, row.name=1,check.names=FALSE, quote="")
	
		edge.file <- paste(wd, "EdgeList_", i, ".txt", sep="")
		write("TF\tTG\ts1", file=edge.file)
		for (u in 1:dim(i.s1)[2]) {
			for (v in 1:dim(i.s1)[1]) {
				u.symbol <- as.character(colnames(i.s1)[u])
				v.symbol <- as.character(rownames(i.s1)[v])
				if (u.symbol != v.symbol) {
					write(paste(u.symbol, v.symbol, i.s1[v,u], sep="\t"), file=edge.file, append=T)
				}
			}
		}
		
		cat("\tfinding important TFs in the main connected component of TRN of group", i, "\n")
		
		g <- getGraph(edge.file, 0.1, FALSE)
		TF.importance.metrics = c("DFC","DCC", "DDC", "DC", "CC", "BC")
		im <- getNodesImportance(g,measurements=TF.importance.metrics, componentType="weak", pathType="all", directed=FALSE, normalized=TRUE)
		
		write.table(im, file=paste(wd, "NodeIm.in.TRN_", i,".LCC.txt", sep=""), col.names=T, row.names=T, sep="\t")
		
		im$TF <- 0
		for (j in 1:dim(im)[1]) {
			idx <- which(rownames(fData(ES)) %in% rownames(im)[j])
			if (length(idx) > 0) {
				if (any(fData(ES)$TF[idx] == 1)) {
					im$TF[j] <- 1
				}
			}
		}
		im <- im[which(im$TF == 1),TF.importance.metrics]
		
		im.r <- as.data.frame(apply(im, 2, function(y) rank(-y, ties.method="min")))
		avg.r <- apply(im.r, 1, mean)
		im.r$AVG.RANK <- as.numeric(rank(avg.r, ties.method="min"))
		
		im.r <- im.r[order(im.r$AVG.RANK),]
		
		im.r <- cbind(TF=rownames(im.r), im.r)
		
		write.table(im.r, file=paste(wd, "TF.ranks.in.TRN_", i,".LCC.txt", sep=""), col.names=T, row.names=T, sep="\t")
		
		cat("\tdone\n")
	}
	
	if (verbose) {
		cat("Sincera: driving force analysis completed \n\n")
	}
}


consensus_maximization <- function(x, alpha=4, epslon=0.1, max.iter=100) {

	cat("sincera: consensus maximization...\n")
	wd <- paste(getwd(), dir.delim, "sincera.consensus.maximization.", getTimestamp(), dir.delim, sep="")
	dir.create(wd)
	
	fs <- x
		
	fs.type <- data.frame(NAME=colnames(fs), TYPE=c(1,2,2,2))

	A <- matrix(0, nrow=dim(fs)[1], ncol=dim(fs)[2]*2)
	rownames(A) = rownames(fs)
	colnames(A) = paste("q", seq(1,dim(A)[2],1), sep="")

	for (j in 0:(dim(fs)[2]-1)) {
		for (i in 1:dim(fs)[1]) {
			A[i,(2*j+1)] <- fs[i, (j+1)]
			A[i,(2*j+2)] <- 1-fs[i, (j+1)]
		}
	}

	Y <- matrix(0, nrow=dim(A)[2], ncol=2)
	rownames(Y) <- colnames(A)
	colnames(Y) <- c("P","N")

	type2.p <- 1
	type2.n <- 0.5

	for (j in 0:(dim(fs)[2]-1) ) {
		if (fs.type$TYPE[j+1] == 1) {
			Y[(2*j+1),] <- c(1,0)
			Y[(2*j+2),] <- c(0, 1)
		} else if (fs.type$TYPE[j+1] == 2) {
			Y[(2*j+1),] <- c(type2.p,1-type2.p)
			Y[(2*j+2),] <- c(type2.n,1-type2.n)
		}
	}

	# start integration
	Q <- matrix(0, nrow=dim(Y)[1], ncol=dim(Y)[2])
	rownames(Q) <- rownames(Y)
	colnames(Q) <- colnames(Y)

	U <- matrix(0, nrow=dim(A)[1], ncol=dim(Y)[2])
	rownames(U) <- rownames(A)
	colnames(U) <- colnames(Y)

	# init U 
	U[,1] <- fs[,2]
	U[,2] <- 1-U[,1]

	iter <- 1

	while (iter < max.iter) {
		iter <- iter + 1
		U.1 <- U
		# update Q
		dv <- matrix(0, nrow=dim(Q)[1], ncol=dim(Q)[1])
		diag(dv) <- apply(A, 2, sum) 
		kv <- matrix(0, nrow=dim(Q)[1], ncol=dim(Q)[1]) 
		diag(kv) <- 1
		Q <- solve(dv + alpha*kv) %*% (t(A) %*% U + (alpha*kv) %*% Y)
		# update U
		dn <- matrix(0, nrow=dim(A)[1], ncol=dim(A)[1])
		diag(dn) <- apply(A, 1, sum)
		U <- solve(dn) %*% A %*% Q
	}

	dU <- cbind(RID=rownames(fs), U)
	dU <- cbind(dU, x)
	dU <- dU[order(dU$P, decreasing=T),]
	
	write.table(dU, file=paste(wd, "CM.U-a", alpha, "-s", max.iter, ".txt", sep=""), sep="\t", col.names=T, row.names=F) 

	cat("sincera: consensus maximization completed\n\n")
}

